import importlib
import logging
import sys
from pathlib import Path
from typing import Dict

import yaml
from pydantic.error_wrappers import ValidationError

from koza.exceptions import MapItemException, NextRowException
from koza.io.writer.jsonl_writer import JSONLWriter
from koza.io.writer.tsv_writer import TSVWriter
from koza.io.writer.writer import KozaWriter
from koza.model.config.source_config import MapFileConfig, OutputFormat
from koza.model.curie_cleaner import CurieCleaner
from koza.model.map_dict import MapDict
from koza.model.source import Source
from koza.model.translation_table import TranslationTable

LOG = logging.getLogger(__name__)


class KozaApp:
    """
    Core koza class that stores the configuration
    and performs source transforms
    """

    def __init__(
        self,
        source: Source,
        translation_table: TranslationTable = None,
        output_dir: str = './output',
        output_format: OutputFormat = OutputFormat('jsonl'),
    ):
        self.source = source
        self.translation_table = translation_table
        self.output_dir = output_dir
        self.output_format = output_format
        self._map_registry: Dict[str, Source] = {}
        self._map_cache: Dict[str, Dict] = {}
        self.curie_cleaner: CurieCleaner = CurieCleaner()
        self.writer: KozaWriter = self._get_writer(
            source.config.name, source.config.node_properties, source.config.edge_properties
        )

        logging.getLogger(__name__)

        if source.config.depends_on is not None:
            for map_file in source.config.depends_on:
                with open(map_file, 'r') as map_file_fh:
                    map_file_config = MapFileConfig(**yaml.safe_load(map_file_fh))
                    map_file_config.transform_code = (
                        str(Path(map_file).parent / Path(map_file).stem) + '.py'
                    )
                self._map_registry[map_file_config.name] = Source(map_file_config)

        self.writer = self._get_writer(
            source.config.name, source.config.node_properties, source.config.edge_properties
        )

        # load the map cache
        for map_file in self._map_registry.values():
            if map_file.config.name not in self._map_cache:
                self._load_map(map_file.config)

    def get_map(self, map_name: str):
        return self._map_cache[map_name]

    def get_row(self, ingest_name: str = None) -> Dict:
        if ingest_name and ingest_name == self.source.config.name:
            return next(self.source)
        elif ingest_name in self._map_registry:
            return next(self._map_registry[ingest_name])
        elif ingest_name:
            raise KeyError(f"{ingest_name} is not the name of the source file or a mapping file")
        else:
            return next(self.source)

    def process_sources(self):
        """
        Transform an entire file using ingest logic in a functionless python file
        where the path to this file is stored in source.transform_code
        or inferred by taking the name and path of the config file and looking for
        a .py file along side it (see constructor)

        Intended for decoupling ingest logic into a configuration like file

        """
        import sys

        parent_path = Path(self.source.config.transform_code).parent
        transform_code = Path(self.source.config.transform_code).stem
        sys.path.append(str(parent_path))
        is_first = True
        transform_module = None

        if self.source.config.transform_mode == 'flat':
            while True:
                try:
                    if is_first:
                        transform_module = importlib.import_module(transform_code)
                        is_first = False
                    else:
                        importlib.reload(transform_module)
                except MapItemException as mie:
                    LOG.warning(f"{str(mie)} not found in map")
                except NextRowException:
                    continue
                except ValidationError as ve:
                    LOG.error(f"Validation error while processing: {self.source.last_row}")
                    raise ve
                except StopIteration:
                    break
        elif self.source.config.transform_mode == 'loop':
            if transform_code not in sys.modules.keys():
                importlib.import_module(transform_code)
            else:
                importlib.reload(importlib.import_module(transform_code))
        else:
            raise NotImplementedError

        # close the writer when the source is done processing
        self.writer.finalize()

        # remove directory from sys.path to prevent name clashes
        sys.path.remove(str(parent_path))

    @staticmethod
    def next_row():
        """
        Breaks out of the facade file and iterates to the next row
        in the file

        Effectively a continue statement
        https://docs.python.org/3/reference/simple_stmts.html#continue
        :return:
        """
        raise NextRowException

    def write(self, *entities):
        self.writer.write(entities)

    def _get_writer(self, name, node_properties, edge_properties):
        if self.output_format == OutputFormat.tsv:
            return TSVWriter(self.output_dir, name, node_properties, edge_properties)

        elif self.output_format == OutputFormat.jsonl:
            return JSONLWriter(self.output_dir, name)

    def _load_map(self, map_file_config: MapFileConfig):
        map_file = Source(map_file_config)

        map = MapDict()

        self._map_cache[map_file_config.name] = map

        transform_code_pth = Path(map_file_config.transform_code)

        if transform_code_pth.exists():
            parent_path = transform_code_pth.parent
            transform_code = transform_code_pth.stem
            sys.path.append(str(parent_path))
            is_first = True
            transform_module = None

            while True:
                try:
                    if is_first:
                        transform_module = importlib.import_module(transform_code)
                        is_first = False
                    else:
                        importlib.reload(transform_module)
                except StopIteration:
                    break
        else:
            key_column = map_file_config.key
            value_columns = map_file_config.values
            for row in map_file:
                map[row[key_column]] = {
                    key: value for key, value in row.items() if key in value_columns
                }



import logging
from dataclasses import dataclass
from typing import Dict, Optional

LOG = logging.getLogger(__name__)


def is_dictionary_bimap(dictionary: Dict[str, str]) -> bool:
    """
    Test if a dictionary is a bimap
    :param dictionary:
    :return: boolean
    """
    is_bimap = True
    all_values = set()
    failed_list = []

    for val in dictionary.values():
        if val not in all_values:
            all_values.add(val)
        else:
            is_bimap = False
            failed_list.append(val)

    if not is_bimap:
        LOG.warning(f"Duplicate values in yaml: {failed_list}")

    return is_bimap


@dataclass(frozen=True)
class TranslationTable:
    """
    Translation table
    """

    global_table: Dict[str, str]
    local_table: Dict[str, str]  # maybe bidict

    def __post_init__(self):
        if not is_dictionary_bimap(self.global_table):
            raise ValueError("Global table is not a bimap")

        if not is_dictionary_bimap(self.local_table):
            raise ValueError("Local table is not a bimap")

    def resolve_term(
        self, word: str, mandatory: Optional[bool] = True, default: Optional[str] = None
    ):
        """
        Resolve a term from a source to its preferred curie

        given a term in some source
        return global[ (local[term] | term) ] || local[term] || (term | default)

        if finding a mapping is not mandatory
        returns x | default on fall through

        This may be generalized further from any mapping
        to a global mapping only; if need be.

        :param word:  the string to find as a key in translation tables
        :param mandatory: boolean to cause failure when no key exists
        :param default: string to return if nothing is found (& not manandatory)
        :return
            value from global translation table,
            or value from local translation table,
            or the query key if finding a value is not mandatory (in this order)
        """

        if word is None:
            raise ValueError("word is required")

        # we may not agree with a remote sources use of a global term we have
        # this provides opportunity for us to override
        if word in self.local_table:
            label = self.local_table[word]
            if label in self.global_table:
                term_id = self.global_table[label]
            else:
                LOG.info("Translated to '%s' but no global term_id for: '%s'", label, word)  #
                term_id = label
        elif word in self.global_table:
            term_id = self.global_table[word]
        else:
            if mandatory:
                raise KeyError("Mapping required for: ", word)
            LOG.warning("We have no translation for: '%s'", word)

            if default is not None:
                term_id = default
            else:
                term_id = word
        return term_id



"""
Testing custom dictionary
"""
import pytest

from koza.exceptions import MapItemException
from koza.model.map_dict import MapDict


def test_custom_dict_exception():
    map_dict = MapDict(foo='bar')
    with pytest.raises(MapItemException):
        map_dict['bad_key']


def test_custom_dict_get_item():
    map_dict = MapDict(foo='bar')
    assert map_dict['foo'] == 'bar'



import pytest
from biolink_model_pydantic.model import Gene, GeneToPhenotypicFeatureAssociation, PhenotypicFeature


@pytest.fixture
def gene2phenotype_entities(mock_koza):
    row = iter(
        [
            {
                'SUBJECT': 'XB-GENE-1000632',
                'SUBJECT_LABEL': 'dctn2',
                'SUBJECT_TAXON': 'NCBITaxon:8364',
                'SUBJECT_TAXON_LABEL': 'Xla',
                'OBJECT': 'XPO:0102358',
                'OBJECT_LABEL': 'abnormal tail morphology',
                'RELATION': 'RO_0002200',
                'RELATION_LABEL': 'has_phenotype',
                'EVIDENCE': '',
                'EVIDENCE_LABEL': '',
                'SOURCE': 'PMID:17112317',
                'IS_DEFINED_BY': '',
                'QUALIFIER': '',
            }
        ]
    )
    return mock_koza('gene-to-phenotype', row, './examples/xenbase/gene-to-phenotype.py')


def test_gene2_phenotype_transform(gene2phenotype_entities):
    assert gene2phenotype_entities
    assert len(gene2phenotype_entities) == 3
    genes = [entity for entity in gene2phenotype_entities if isinstance(entity, Gene)]
    phenotypes = [
        entity for entity in gene2phenotype_entities if isinstance(entity, PhenotypicFeature)
    ]
    associations = [
        entity
        for entity in gene2phenotype_entities
        if isinstance(entity, GeneToPhenotypicFeatureAssociation)
    ]
    assert len(genes) == 1
    assert len(phenotypes) == 1
    assert len(associations) == 1


def test_gene2_phenotype_transform_publications(gene2phenotype_entities):
    associations = [
        entity
        for entity in gene2phenotype_entities
        if isinstance(entity, GeneToPhenotypicFeatureAssociation)
    ]
    assert 'PMID:17112317' in associations[0].publications



#!/usr/bin/env python3
"""
CLI interface for Koza
"""

import logging
from pathlib import Path

import typer

from koza.cli_runner import transform_source, validate_file
from koza.model.config.source_config import CompressionType, FormatType, OutputFormat

typer_app = typer.Typer()


logging.basicConfig()
LOG = logging.getLogger(__name__)


@typer_app.command()
def transform(
    source: str = typer.Option(..., help="Source metadata file"),
    output_dir: str = typer.Option('./output', help="Path to output directory"),
    output_format: OutputFormat = typer.Option("jsonl", help="Output format"),
    global_table: str = typer.Option(None, help="Path to global translation table"),
    local_table: str = typer.Option(None, help="Path to local translation table"),
    quiet: bool = False,
    debug: bool = False,
):
    """
    Transform a source file
    """
    _set_log_level(quiet, debug)

    output_path = Path(output_dir)

    if output_path.exists() and not output_path.is_dir():
        raise NotADirectoryError(f"{output_dir} is not a directory")
    elif not output_path.exists():
        output_path.mkdir(parents=True)

    transform_source(source, output_dir, output_format, global_table, local_table)


@typer_app.command()
def validate(
    file: str = typer.Option(..., help="Path or url to the source file"),
    format: FormatType = FormatType.csv,
    delimiter: str = ',',
    header_delimiter: str = None,
    compression: CompressionType = None,
    skip_lines: int = 0,
    skip_blank_lines: bool = True,
):
    """
    Validate a source file

    Given a file and configuration checks that the file is valid, ie
    format is as expected (tsv, json), required columns/fields are there
    """
    _set_log_level(debug=True)

    validate_file(
        file, format, delimiter, header_delimiter, compression, skip_lines, skip_blank_lines
    )


# @typer_app.command()
# def create():
#    """
#    TODO
#    Create a new koza project
#    """


def _set_log_level(quiet: bool = False, debug: bool = False):
    if quiet:
        logging.getLogger().setLevel(logging.WARNING)
    elif debug:
        logging.getLogger().setLevel(logging.DEBUG)
    else:
        logging.getLogger().setLevel(logging.INFO)


if __name__ == "__main__":
    typer_app()



"""
Testing for row filtering

"""
import pytest

from koza.model.config.source_config import ColumnFilter, FilterCode, FilterInclusion
from koza.row_filter import RowFilter


@pytest.mark.parametrize(
    "column, inclusion, code, value, result",
    [
        ('a', 'include', 'eq', 0.3, True),
        ('a', 'exclude', 'eq', 0.3, False),
        ('a', 'include', 'ge', 0.3, True),
        ('a', 'include', 'le', 0.3, True),
        ('a', 'include', 'ne', 0.3, False),
        ('a', 'include', 'ne', 0.4, True),
        ('a', 'include', 'lt', 0.4, True),
        ('a', 'include', 'gt', 0.4, False),
        ('a', 'include', 'gt', 0.2, True),
        ('a', 'exclude', 'gt', 0.4, True),
        ('a', 'exclude', 'gt', 0.2, False),
        ('b', 'include', 'eq', 10, True),
        ('b', 'include', 'ge', 10, True),
        ('b', 'include', 'le', 10, True),
        ('b', 'include', 'gt', 9, True),
        ('b', 'include', 'gt', 11, False),
        ('b', 'include', 'lt', 11, True),
        ('b', 'include', 'lt', 9, False),
        ('c', 'include', 'eq', 'llama', True),
        ('c', 'include', 'ne', 'alpaca', True),
        ('c', 'include', 'eq', 'alpaca', False),
        ('c', 'include', 'in', ['llama', 'alpaca'], True),
        ('c', 'include', 'in', ['condor', 'alpaca'], False),
    ],
)
def test_filter(column, inclusion, code, value, result):
    row = {'a': 0.3, 'b': 10, 'c': 'llama'}
    column_filter = ColumnFilter(
        column=column,
        inclusion=FilterInclusion(inclusion),
        filter_code=FilterCode(code),
        value=value,
    )

    rf = RowFilter([column_filter])

    assert rf.include_row(row) is result


@pytest.mark.parametrize(
    "column_filters, result",
    [
        (
            [
                ColumnFilter(
                    column='a',
                    inclusion=FilterInclusion('include'),
                    filter_code=FilterCode('lt'),
                    value=0.4,
                ),
                ColumnFilter(
                    column='a',
                    inclusion=FilterInclusion('include'),
                    filter_code=FilterCode('gt'),
                    value=0.1,
                ),
            ],
            True,
        ),
        (
            [
                ColumnFilter(
                    column='a',
                    inclusion=FilterInclusion('include'),
                    filter_code=FilterCode('lt'),
                    value=0.4,
                ),
                ColumnFilter(
                    column='a',
                    inclusion=FilterInclusion('exclude'),
                    filter_code=FilterCode('gt'),
                    value=0.4,
                ),
            ],
            True,
        ),
        (
            [
                ColumnFilter(
                    column='a',
                    inclusion=FilterInclusion('include'),
                    filter_code=FilterCode('in'),
                    value=[0.2, 0.3, 0.4],
                ),
                ColumnFilter(
                    column='b',
                    inclusion=FilterInclusion('exclude'),
                    filter_code=FilterCode('lt'),
                    value=9,
                ),
            ],
            True,
        ),
        (
            [
                ColumnFilter(
                    column='a',
                    inclusion=FilterInclusion('include'),
                    filter_code=FilterCode('in'),
                    value=[0.2, 0.3, 0.4],
                ),
                ColumnFilter(
                    column='b',
                    inclusion=FilterInclusion('exclude'),
                    filter_code=FilterCode('gt'),
                    value=9,
                ),
            ],
            False,
        ),
    ],
)
def test_multiple_filters(column_filters, result):
    row = {'a': 0.3, 'b': 10, 'c': 'llama'}

    rf = RowFilter(column_filters)

    assert rf.include_row(row) == result


def test_empty_filters():
    row = {'a': 0.3, 'b': 10, 'c': 'llama'}

    rf = RowFilter()

    assert rf.include_row(row)



from abc import ABC, abstractmethod
from typing import Iterable


class KozaWriter(ABC):
    """
    An abstract base class for all koza writers

    # @abstractmethod
    # def writeheader(self) -> Optional[int]:
    #     pass
    """

    @abstractmethod
    def write(self, entities: Iterable):
        pass

    @abstractmethod
    def finalize(self):
        pass



from dataclasses import asdict
from typing import Iterable

from biolink_model_pydantic.model import Association, NamedThing


class KGXConverter:
    """
    Converts the biolink model to the KGX format, which splits
    data into nodes (Entity) and edges (Association)

    This format is designed around labelled property graphs
    that allow for scalar and lists as properties (Neo4J)

    https://github.com/biolink/kgx/blob/master/specification/kgx-format.md

    """

    def convert(self, entities: Iterable) -> (dict, dict):

        nodes = []
        edges = []

        for entity in entities:
            if isinstance(entity, NamedThing):
                nodes.append(self.convert_node(entity))
            elif isinstance(entity, Association):
                edges.append(self.convert_association(entity))
            else:
                raise ValueError(
                    "Can only convert NamedThing or Association entities to KGX compatible dictionaries"
                )

        return nodes, edges

    def convert_node(self, node: NamedThing) -> dict:
        return asdict(node)

    def convert_association(self, association: Association) -> dict:
        return asdict(association)



import logging
import uuid

from biolink_model_pydantic.model import (
    Gene,
    NamedThingToInformationContentEntityAssociation,
    Predicate,
    Publication,
)

from koza.cli_runner import koza_app

LOG = logging.getLogger(__name__)

source_name = 'gene-literature'

row = koza_app.get_row(source_name)
genepage2gene = koza_app.get_map('genepage-2-gene')

entities = []

gene_pages = row['gene_pages']

publication = Publication(
    id='PMID:' + row['pmid'], type=koza_app.translation_table.resolve_term("publication")
)

entities.append(publication)

for gene_page in gene_pages.split(','):
    gene_page_id = gene_page.split(' ')[0]
    try:
        gene_ids = map(lambda id: f"Xenbase:{id}", list(genepage2gene[gene_page_id].values()))
    except KeyError:
        LOG.debug(f"Could not locate genepage_id: {gene_page_id} in row {row}")
        continue
    for gene_id in gene_ids:
        gene = Gene(id=gene_id)

        entities.append(gene)

        association = NamedThingToInformationContentEntityAssociation(
            id="uuid:" + str(uuid.uuid1()),
            subject=gene.id,
            predicate=Predicate.mentions,
            object=publication.id,
            relation="IAO:0000142",  # Mentions
        )

        entities.append(association)

koza_app.write(*entities)



import json
import os
from dataclasses import asdict
from typing import List

from biolink_model_pydantic.model import Association, Entity, NamedThing

from koza.io.writer.writer import KozaWriter


class JSONLWriter(KozaWriter):
    def __init__(self, output_dir: str, source_name: str):
        self.output_dir = output_dir
        self.source_name = source_name

        os.makedirs(output_dir, exist_ok=True)

        self.nodes_file = open(f"{output_dir}/{source_name}_nodes.jsonl", "w")
        self.edges_file = open(f"{output_dir}/{source_name}_edges.jsonl", "w")

    def write(self, entities: List[Entity]):
        for entity in entities:
            if isinstance(entity, NamedThing):
                node = json.dumps(asdict(entity), ensure_ascii=False)
                self.nodes_file.write(node + '\n')
            elif isinstance(entity, Association):
                edge = json.dumps(asdict(entity), ensure_ascii=False)
                self.edges_file.write(edge + '\n')
            else:
                raise ValueError("Can only write NamedThing or Association entities")

    def finalize(self):
        self.nodes_file.close()
        self.edges_file.close()



class PydanticConfig:
    """
    Pydantic config
    https://pydantic-docs.helpmanual.io/usage/model_config/
    """

    validate_assignment = True
    validate_all = True
    underscore_attrs_are_private = True
    extra = 'forbid'



from typing import Iterable, List

from kgx.sink.tsv_sink import TsvSink

from koza.converter.kgx_converter import KGXConverter
from koza.io.writer.writer import KozaWriter


class TSVWriter(KozaWriter):
    def __init__(
        self, output_dir, source_name: str, node_properties: List[str], edge_properties: List[str]
    ):
        self.output_dir = output_dir
        self.source_name = source_name
        self.converter = KGXConverter()

        if node_properties and edge_properties:
            self.sink = TsvSink(
                f"{output_dir}/{source_name}",
                "tsv",
                node_properties=node_properties,
                edge_properties=edge_properties,
            )
        else:  # allow the TsvSink to set defaults if no properties are specified
            self.sink = TsvSink(f"{output_dir}/{source_name}", "tsv")

    def write(self, entities: Iterable):

        (nodes, edges) = self.converter.convert(entities)

        for node in nodes:
            self.sink.write_node(node)

        for edge in edges:
            self.sink.write_edge(edge)

    def finalize(self):
        self.sink.finalize()



import gzip
from pathlib import Path

import pytest

from koza.io.reader.jsonl_reader import JSONLReader

test_zfin = (
    Path(__file__).parent.parent / 'resources' / 'source-files' / 'ZFIN_PHENOTYPE_0.jsonl.gz'
)


def test_normal_case():
    with gzip.open(test_zfin, 'rt') as zfin:
        jsonl_reader = JSONLReader(zfin)
        row = next(jsonl_reader)
        assert len(row) == 6
        assert row['objectId'] == 'ZFIN:ZDB-GENE-011026-1'


def test_required_property():
    with gzip.open(test_zfin, 'rt') as zfin:
        jsonl_reader = JSONLReader(zfin, ['objectId'])
        for row in jsonl_reader:
            # assert len(row) == 1  # removed subsetter
            assert 'objectId' in row


def test_missing_req_property_raises_exception():
    with gzip.open(test_zfin, 'rt') as zfin:
        jsonl_reader = JSONLReader(zfin, ['objectId', 'foobar'])
        with pytest.raises(ValueError):
            next(jsonl_reader)



import logging

from koza.cli_runner import koza_app
from koza.converter.biolink_converter import gpi2gene

LOG = logging.getLogger(__name__)

row = koza_app.get_row('gene-information')

row['DB_Object_ID'] = "Xenbase:" + row['DB_Object_ID']

gene = gpi2gene(row)

koza_app.write(gene)



import uuid

from biolink_model_pydantic.model import (
    Gene,
    GeneToPhenotypicFeatureAssociation,
    PhenotypicFeature,
    Predicate,
)

from koza.cli_runner import koza_app

row = koza_app.get_row()

gene = Gene(id='Xenbase:' + row['SUBJECT'])

phenotype = PhenotypicFeature(id=row['OBJECT'])

association = GeneToPhenotypicFeatureAssociation(
    id="uuid:" + str(uuid.uuid1()),
    subject=gene.id,
    predicate=Predicate.has_phenotype,
    object=phenotype.id,
    relation=row['RELATION'].replace('_', ':'),
)

if row['SOURCE']:
    association.publications = [row['SOURCE']]

koza_app.write(gene, association, phenotype)



"""
Module for managing koza runs through the CLI
"""

import logging
from pathlib import Path
from typing import Optional

import yaml

from koza.app import KozaApp
from koza.io.reader.csv_reader import CSVReader
from koza.io.reader.json_reader import JSONReader
from koza.io.reader.jsonl_reader import JSONLReader
from koza.io.utils import open_resource
from koza.model.config.source_config import (
    CompressionType,
    FormatType,
    OutputFormat,
    PrimaryFileConfig,
)
from koza.model.source import Source
from koza.model.translation_table import TranslationTable

LOG = logging.getLogger(__name__)

koza_app = None


def set_koza(koza: KozaApp):
    global koza_app
    koza_app = koza


def set_koza_app(
    source: Source,
    translation_table: TranslationTable = None,
    output_dir: str = './output',
    output_format: OutputFormat = OutputFormat('jsonl'),
) -> KozaApp:
    """
    Setter for singleton koza app object
    """
    global koza_app

    koza_app = KozaApp(source, translation_table, output_dir, output_format)
    return koza_app


def get_koza_app() -> Optional[KozaApp]:
    """
    Getter for singleton koza app object
    """
    return koza_app


def validate_file(
    file: str,
    format: FormatType = FormatType.csv,
    delimiter: str = ',',
    header_delimiter: str = None,
    compression: CompressionType = None,
    skip_lines: int = 0,
    skip_blank_lines: bool = True,
):
    """
    Runs a file through one of the Koza readers
    For csv files will return header and row length

    For json and jsonl just validates them
    """

    with open_resource(file, compression) as resource_io:

        if format == FormatType.csv:
            reader = CSVReader(
                resource_io,
                delimiter=delimiter,
                header_delimiter=header_delimiter,
                skip_lines=skip_lines,
                skip_blank_lines=skip_blank_lines,
            )
        elif format == FormatType.jsonl:
            reader = JSONLReader(resource_io)
        elif format == FormatType.json:
            reader = JSONReader(resource_io)
        else:
            raise ValueError

        for _ in reader:
            pass


def get_translation_table(global_table: str = None, local_table: str = None) -> TranslationTable:
    """
    Create a translation table object from two file paths
    :param global_table: str, path to global table yaml
    :param local_table: str, path to local table yaml
    :return: TranslationTable
    """

    global_tt = {}
    local_tt = {}

    if not global_table:
        if local_table:
            raise ValueError("Local table without a global table not allowed")
        else:
            LOG.info("No global table used for transform")
    else:

        with open(global_table, 'r') as global_tt_fh:
            global_tt = yaml.safe_load(global_tt_fh)

        if local_table:
            with open(local_table, 'r') as local_tt_fh:
                local_tt = yaml.safe_load(local_tt_fh)
        else:
            LOG.info("No local table used for transform")

    return TranslationTable(global_tt, local_tt)


def transform_source(
    source: str,
    output_dir: str,
    output_format: OutputFormat,
    global_table: str = None,
    local_table: str = None,
):

    translation_table = get_translation_table(global_table, local_table)

    with open(source, 'r') as source_fh:
        source_config = PrimaryFileConfig(**yaml.safe_load(source_fh))
        if not source_config.name:
            source_config.name = Path(source).stem

        if not source_config.transform_code:
            # look for it alongside the source conf as a .py file
            source_config.transform_code = str(Path(source).parent / Path(source).stem) + '.py'

        koza_source = Source(source_config)

        koza_app = set_koza_app(koza_source, translation_table, output_dir, output_format)
        koza_app.process_sources()



from typing import Any, Dict, Iterator, List, Optional, Union

import yaml

from koza.io.reader.csv_reader import CSVReader
from koza.io.reader.json_reader import JSONReader
from koza.io.reader.jsonl_reader import JSONLReader
from koza.io.utils import open_resource
from koza.model.config.source_config import MapFileConfig, PrimaryFileConfig, SourceConfig
from koza.row_filter import RowFilter


class Source:
    """
    Source class for files and maps

    Source is an iterator that provides a layer of abstraction over file types
    and adds filter support to the readers in io.reader

    config: Source config
    reader: An iterator that takes in an IO[str] as its first argument
    and yields a dictionary
    """

    def __init__(
        self,
        config: Union[PrimaryFileConfig, MapFileConfig, str],
    ):

        self.config = config
        self._filter = RowFilter(config.filters)
        self._reader = None
        self._readers: List = []
        self.last_row: Optional[Dict] = None

        if not isinstance(config, SourceConfig):
            # Check to see if it's a file path
            with open(config, 'r') as source_file_fh:
                self.config = PrimaryFileConfig(**yaml.safe_load(source_file_fh))
        else:
            # TODO better error handling
            self.config = config

        for file in config.files:
            resource_io = open_resource(file, config.compression)
            if self.config.format == 'csv':
                self._readers.append(
                    CSVReader(
                        resource_io,
                        name=config.name,
                        field_type_map=config.field_type_map,
                        delimiter=config.delimiter,
                        header_delimiter=config.header_delimiter,
                        has_header=config.has_header,
                        skip_lines=config.skip_lines,
                    )
                )
            elif self.config.format == 'jsonl':
                self._readers.append(
                    JSONLReader(
                        resource_io,
                        name=config.name,
                        required_properties=config.required_properties,
                    )
                )
            elif self.config.format == 'json' or self.config.format == 'yaml':
                self._readers.append(
                    JSONReader(
                        resource_io,
                        name=config.name,
                        json_path=config.json_path,
                        required_properties=config.required_properties,
                        is_yaml=(self.config.format == 'yaml'),
                    )
                )
            else:
                raise ValueError(f"File type {format} not supported")

    def __iter__(self) -> Iterator:
        return self

    def __next__(self) -> Dict[str, Any]:
        if self._reader is None:
            self._reader = self._readers.pop()
        try:
            row = self._get_row()
        except StopIteration as si:
            if len(self._readers) == 0:
                raise si
            else:
                self._reader = self._readers.pop()
                row = self._get_row()
        return row

    def _get_row(self):
        if self._filter:
            row = next(self._reader)
            while not self._filter.include_row(row):
                # TODO log filtered out lines
                row = next(self._reader)
        else:
            row = next(self._reader)
        # Retain the most recent row so that it can be logged alongside validation errors
        self.last_row = row
        return row



from operator import eq, ge, gt, le, lt, ne
from typing import List

from koza.model.config.source_config import ColumnFilter, FilterInclusion


class RowFilter:
    """
    A Filter class that is initialized with a List of column filters, each specifying a column, an operator and a value
    """

    def __init__(self, filters: List[ColumnFilter] = None):
        """
        :param filters: A collection of Filters to be applied
        """
        self.filters = filters
        self.operators = {
            'gt': gt,
            'ge': ge,
            'lt': lt,
            'le': le,
            'eq': eq,
            'ne': ne,
            'in': self.inlist,  # not using operator.contains because the it expects opposite argument order
        }

    def include_row(self, row) -> bool:
        """
        :param row: A dictionary representing a single row
        :return: bool for whether the row should be included
        """

        if not self.filters:
            return True

        include_row = True

        column_filter: ColumnFilter
        for column_filter in self.filters:
            # None can't be greater, less than or equal to any specified value, right?
            if row.get(column_filter.column) is None:
                return False

            include = column_filter.inclusion == FilterInclusion('include')
            exclude = column_filter.inclusion == FilterInclusion('exclude')

            comparison_method = self.operators.get(column_filter.filter_code)
            comparison_match = comparison_method(row.get(column_filter.column), column_filter.value)

            if (include and not comparison_match) or (exclude and comparison_match):
                include_row = False

        return include_row

    def inlist(self, column_value, filter_value):
        return column_value in filter_value



import uuid

from biolink_model_pydantic.model import Gene, PairwiseGeneToGeneInteraction, Predicate

from koza.cli_runner import koza_app

row = koza_app.get_row()
entrez_2_string = koza_app.get_map('custom_entrez_2_string')

gene_a = Gene(id='NCBIGene:' + entrez_2_string[row['protein1']]['entrez'])
gene_b = Gene(id='NCBIGene:' + entrez_2_string[row['protein2']]['entrez'])

pairwise_gene_to_gene_interaction = PairwiseGeneToGeneInteraction(
    id="uuid:" + str(uuid.uuid1()),
    subject=gene_a.id,
    object=gene_b.id,
    predicate=Predicate.interacts_with,
    relation=koza_app.translation_table.global_table['interacts with'],
)

koza_app.write(gene_a, gene_b, pairwise_gene_to_gene_interaction)



import json
import logging
from typing import IO, Any, Dict, Iterator, List

LOG = logging.getLogger(__name__)


class JSONLReader:
    """
    A simple JSON lines reader that optionally returns a subset of
    configured properties
    """

    def __init__(
        self, io_str: IO[str], required_properties: List[str] = None, name: str = 'jsonl file'
    ):
        """
        :param io_str: Any IO stream that yields a string
                       See https://docs.python.org/3/library/io.html#io.IOBase
        :param required_properties: List of required top level properties
        :param name: todo
        """
        self.io_str = io_str
        self.required_properties = required_properties
        self.line_num = 0
        self.name = name

    def __iter__(self) -> Iterator:
        return self

    def __next__(self) -> Dict[str, Any]:
        next_line = self.io_str.readline()
        if not next_line:
            LOG.info(f"Finished processing {self.line_num} lines")
            raise StopIteration
        self.line_num += 1
        json_obj = json.loads(next_line)

        if self.required_properties:
            if not set(json_obj.keys()) >= set(self.required_properties):
                # TODO - have koza runner handle this exception
                # based on some configuration? similar to
                # on_map_error
                raise ValueError(
                    f"Configured properties missing in source file "
                    f"{set(self.required_properties) - set(json_obj.keys())}"
                )
            # If we want to turn this into a subsetter
            # json_obj = {key: json_obj[key] for key in json_obj.keys() if key in self.required_properties}

        return json_obj



#!/usr/bin/env python3
"""
Set of functions to manage input and output
"""
import gzip
import tempfile
from io import TextIOWrapper
from os import PathLike
from pathlib import Path, PosixPath
from typing import IO, Union

import requests

from koza.model.config.source_config import CompressionType


def open_resource(resource: Union[str, PathLike], compression: CompressionType = None) -> IO[str]:
    """
    A generic function for opening a local or remote file

    On remote files - files are written to a temporary file, returned as an IO[str]
    and then deleted upon closing.  Users of this lib should be encouraged to
    fetch remote files and store them locally using a more specialized tool
    wget --timestamping with gmake works great see
    https://github.com/monarch-initiative/DipperCache

    Currently no plans to support FTP, but note
    that requests does not support FTP (consider ftplib or urllib.request)

    :param resource: str or PathLike - local filepath or remote resource
    :param compression: str or PathLike - compression type
    :return: str, next line in resource

    """
    if isinstance(resource, PosixPath) and Path(resource).exists():
        if compression is None:
            # Try gzip first
            try:
                file = gzip.open(resource, 'rt')
                file.read(1)
                file.seek(0)

            except OSError:
                file = open(resource, 'r')
        elif compression == CompressionType.gzip:
            file = gzip.open(resource, 'rt')
        else:
            file = open(resource, 'r')

        return file

    elif isinstance(resource, str) and resource.startswith('http'):
        tmp_file = tempfile.TemporaryFile('w+b')
        request = requests.get(resource)
        if request.status_code != 200:
            raise ValueError(f"Remote file returned {request.status_code}: {request.text}")
        tmp_file.write(request.content)
        request.close()  # not sure this is needed
        tmp_file.seek(0)
        if resource.endswith('gz') or compression == CompressionType.gzip:
            # This should be more robust, either check headers
            # or use https://github.com/ahupp/python-magic
            remote_file = gzip.open(tmp_file, 'rt')
            return remote_file

        else:
            return TextIOWrapper(tmp_file)

    else:
        raise ValueError(f"Cannot open local or remote file: {resource}")



from koza.exceptions import MapItemException


class MapDict(dict):
    """
    A custom dictionary that raises a special KeyError exception
    MapItemException
    """

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError as key_error:
            raise MapItemException(*key_error.args)



from koza.cli_runner import koza_app

source_name = "custom_entrez_2_string"

row = koza_app.get_row(source_name)

map = koza_app.get_map(source_name)

entry = dict()

entry["entrez"] = row["entrez"]

map[row["STRING"]] = entry



import re
import uuid

from biolink_model_pydantic.model import PairwiseGeneToGeneInteraction, Predicate, Protein

from koza.cli_runner import koza_app

for row in koza_app.source:

    protein_a = Protein(id='ENSEMBL:' + re.sub(r'\d+\.', '', row['protein1']))
    protein_b = Protein(id='ENSEMBL:' + re.sub(r'\d+\.', '', row['protein2']))

    pairwise_gene_to_gene_interaction = PairwiseGeneToGeneInteraction(
        id="uuid:" + str(uuid.uuid1()),
        subject=protein_a.id,
        object=protein_b.id,
        predicate=Predicate.interacts_with,
        relation=koza_app.translation_table.global_table['interacts with'],
    )

    koza_app.write(protein_a, protein_b, pairwise_gene_to_gene_interaction)
